from typing import List, TypeVar, Generic
from abc import ABC, abstractmethod
from contextlib import AsyncExitStack

from vespucci.base import Edge


N = TypeVar("N")


class Store(Generic[N], ABC):
    @abstractmethod
    async def save(self, edge: Edge[N]) -> None:
        pass

    @abstractmethod
    # IMPROVE: The return type should be Optional[List[N]], with None indicating
    # that the node is not in the graph store. However, this makes implementing
    # clients more complicated. Now, returning an empty list means that `src` is
    # not in the store, thereby losing the distinction between not being in the
    # store, and having no suggestions.
    async def load(self, src: N) -> List[N]:
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def enter(self, stack: AsyncExitStack):
        return await stack.enter_async_context(self)



import asyncio
from uuid import uuid4
import logging
from contextlib import AsyncExitStack
from typing import Generic, Set, TypeVar, Optional

from vespucci.oracle import Oracle
from vespucci.store import Store
from vespucci.base import Tag

log = logging.getLogger(__name__)


N = TypeVar("N")


class Crawler(Generic[N]):
    _visited: Set[N]
    _LOG_FREQUENCY = 100

    def __init__(
        self, oracle: Oracle[N], store: Store[N], tag: Optional[Tag] = None, capacity: int = 10
    ):
        self._oracle = oracle
        self._store = store
        self._tag = tag or str(uuid4())
        self._capacity = capacity
        self._process_task = None
        self._visited = set()

    async def seed(self, node: N, rank: int):
        await self._oracle.seed(self._tag, node, rank)
        self._visited.add(node)

    async def _process(self):
        n_edges = 0
        async for (edge, rank) in self._oracle.reveal(self._tag):
            log.debug("Crawler %s found edge %s of rank %d", self._tag, edge, rank)
            await self._store.save(edge)
            if edge.dst is not None:
                n_edges += 1
                if n_edges % self._LOG_FREQUENCY == 0:
                    log.info("Crawler %s traversed %d edges", self._tag, n_edges)

            if rank > 1 and edge.dst is not None and edge.dst not in self._visited:
                dsts = await self._store.load(edge.dst)
                if dsts:
                    # The neighborhood of edge.dst was already in the graph store
                    for dst in dsts:
                        await self.seed(dst, rank - 1)
                else:
                    await self.seed(edge.dst, rank - 1)
            else:
                in_process = self._oracle.in_process(self._tag)
                log.debug("Crawler %s has %d nodes in process", self._tag, in_process)
                if in_process <= 0:
                    log.info("Crawler %s traversed %d edges", self._tag, n_edges)
                    return

    def run(self):
        return self._process_task

    async def __aenter__(self):
        self._oracle.register(self._tag, self._capacity)
        self._process_task = asyncio.create_task(self._process())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self._process_task.done():
            self._process_task.cancel()
        self._oracle.unregister(self._tag)

    async def enter(self, stack: AsyncExitStack):
        return await stack.enter_async_context(self)



from collections import Counter
import asyncio
from asyncio import Queue, create_task
from typing import Mapping, Hashable, TypeVar, Generic, Tuple, MutableMapping
from vespucci.base import Tag


class _Merged(Queue):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._tasks = []

    def stop(self):
        for task in self._tasks:
            task.cancel()

    def count(self, tag: Tag) -> int:
        pass


K = TypeVar("K", bound=Hashable)
T = TypeVar("T")


class Combined(Generic[K, T]):
    _qsize: MutableMapping[K, int]

    def __init__(self, queue: Queue):
        self._queue = queue
        self._qsize = Counter()

    async def get(self) -> Tuple[K, T]:
        (key, item) = await self._queue.get()
        self._qsize[key] -= 1
        if self._qsize[key] == 0:
            del self._qsize[key]
        return (key, item)

    async def put(self, key: K, item: T):
        self._qsize[key] += 1
        return self._queue.put((key, item))

    async def put_nowait(self, key: K, item: T):
        self._qsize[key] += 1
        return self._queue.put_nowait((key, item))

    def qsize(self, key: K) -> int:
        return self._qsize[key]


async def merge(inqs: Mapping[Hashable, Queue], **kwargs) -> Queue:
    async def move(tag, inq, outq):
        try:
            while True:
                item = await inq.get()
                await outq.put((tag, item))
        except asyncio.CancelledError:
            return

    outq = _Merged(**kwargs)

    for (tag, inq) in inqs.items():
        # pylint: disable=protected-access
        outq._tasks.append(create_task(move(tag, inq, outq)))

    return outq



from dataclasses import dataclass
from typing import Generic, TypeVar, Hashable


N = TypeVar("N")


@dataclass(frozen=True, eq=True)
class Edge(Generic[N]):
    src: N
    dst: N


Tag = Hashable


def arc_fn(neighbor_fn):
    async def fn(node):
        nodes = await neighbor_fn(node)
        return [Edge(src=node, dst=n) for n in nodes]

    return fn



import logging
import asyncio
import time


log = logging.getLogger(__name__)


class TokenBucket:
    def __init__(self, fn, rate: float, max_tokens: float):
        self._fn = fn
        self._rate = rate
        self._max_tokens = max_tokens
        self._updated_at = time.monotonic()
        self._tokens = max_tokens
        self._start = time.monotonic()

    async def __call__(self, *args, **kwargs):
        await self.wait_for_token()
        return await self._fn(*args, **kwargs)

    async def wait_for_token(self):
        while self._tokens < 1:
            delay = self.add_new_tokens()
            await asyncio.sleep(delay)
        self._tokens -= 1

    def add_new_tokens(self):
        now = time.monotonic()
        time_since_update = now - self._updated_at
        new_tokens = time_since_update * self._rate
        self._updated_at = now
        self._tokens = min(self._tokens + new_tokens, self._max_tokens)
        if self._tokens >= 1:
            return 0
        delay = (1 - self._tokens) / self._rate
        log.debug("Delaying call for %.02f seconds", delay)
        return delay



# pylint: disable=unused-import
from vespucci.crawler import Crawler
from vespucci.oracle import Oracle
from vespucci.base import Edge, arc_fn
from vespucci.store import Store
from vespucci.utils import merge
from vespucci.token_bucket import TokenBucket
from vespucci.store_stream import StoreStream, Record



from concurrent.futures import Executor
from typing import List, TextIO, TypeVar, Mapping, Any, Iterator, Callable
import csv
import asyncio

from vespucci.base import Edge
from vespucci.store import Store


N = TypeVar("N")

Record = Mapping[str, Any]


class StoreStream(Store[N]):
    def __init__(
        self,
        executor: Executor,
        stream: TextIO,
        records: Callable[[Edge[N]], Iterator[Record]],
    ):
        self._executor = executor
        self._stream = stream
        self._records = records
        self._writer = None

    async def save(self, edge: Edge[N]) -> None:
        def write(record):
            if self._writer is None:
                self._writer = csv.DictWriter(self._stream, fieldnames=record.keys())
                self._writer.writeheader()
            self._writer.writerow(record)
            self._stream.flush()

        for record in self._records(edge):
            await asyncio.get_running_loop().run_in_executor(self._executor, write, record)

    async def load(self, src: N) -> List[N]:
        return []



import asyncio
import logging
from asyncio.queues import LifoQueue, Queue  # pylint: disable=ungrouped-imports
from contextlib import AsyncExitStack
from typing import AsyncGenerator, Callable, Generic, Tuple, TypeVar, MutableMapping

from vespucci.utils import Combined
from vespucci.base import Edge, Tag

log = logging.getLogger(__name__)


N = TypeVar("N")


class Oracle(Generic[N]):

    _seeds: Combined[Tag, Tuple[N, int]]
    _revealed: MutableMapping[Tag, Queue]
    _fn_calls: MutableMapping[Tag, int]

    def __init__(self, fn: Callable[[N], Edge[N]], capacity: int = 100):
        self._fn = fn
        self._capacity = capacity
        self._seeds = Combined(LifoQueue(maxsize=self._capacity))
        self._revealed = {}
        self._fn_calls = {}
        self._process_task = None
        self._merged = None

    async def enter(self, stack: AsyncExitStack):
        return await stack.enter_async_context(self)

    async def seed(self, tag: Tag, node: N, rank: int):
        await self._seeds.put_nowait(tag, (node, rank))

    async def reveal(self, tag: Tag) -> AsyncGenerator[Tuple[Edge[N], int], None]:
        while True:
            (edge, rank) = await self._revealed[tag].get()
            yield (edge, rank)

    def in_process(self, tag: Tag) -> int:
        return self._seeds.qsize(tag) + self._revealed[tag].qsize() + self._fn_calls[tag]

    def register(self, tag: Tag, capacity: int = 0) -> None:
        if tag not in self._revealed:
            self._revealed[tag] = Queue(maxsize=capacity)
            self._fn_calls[tag] = 0

    def unregister(self, tag) -> None:
        del self._fn_calls[tag]
        del self._revealed[tag]

    async def _process(self):
        while True:
            (tag, (node, rank)) = await self._seeds.get()
            # TODO: Deal with self._fn(node) raising exceptions
            self._fn_calls[tag] += 1
            edges = [e for e in await self._fn(node) if e.src != e.dst]
            if edges:
                for edge in edges:
                    await self._revealed[tag].put((edge, rank))
            else:
                # Pass control to a client waiting on a next edge to reveal
                await self._revealed[tag].put((Edge(node, None), rank))
            self._fn_calls[tag] -= 1

    async def __aenter__(self):
        self._process_task = asyncio.create_task(self._process())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._process_task, "cancel"):
            self._process_task.cancel()
